from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync, SocketKind


class Teamspeak3(ProtocolBase):
    """Teamspeak 3 Protocol"""
    full_name = 'Teamspeak 3 Protocol'

    def __init__(self, host: str, port: int, voice_port: int, timeout: float = 5):
        super().__init__(host, port, timeout)
        self._voice_port = voice_port

    async def get_info(self):
        response = await self.__send_and_receive(b'serverinfo')
        return self.__parse_kvs(response)

    async def get_clients(self):
        response = await self.__send_and_receive(b'clientlist')
        return self.__parse_rows(response)

    async def get_channels(self):
        response = await self.__send_and_receive(b'channellist -topic')
        return self.__parse_rows(response)

    async def __send_and_receive(self, data: bytes):
        with SocketAsync(SocketKind.SOCK_STREAM) as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # b'TS3\n\rWelcome to the TeamSpeak 3 ServerQuery interface,
            # type "help" for a list of commands and "help <command>" for information on a specific command.\n\r'
            await sock.recv()

            # b'error id=0 msg=ok\n\r'
            sock.send(f'use port={self._voice_port}\n'.encode())
            await sock.recv()

            sock.send(data + b'\x0A')
            response = await sock.recv()

        return response[:-21]

    def __parse_rows(self, response: bytes):
        return [self.__parse_kvs(row) for row in response.split(b'|')]

    def __parse_kvs(self, response: bytes):
        kvs = {}

        for kv in response.split(b' '):
            index = kv.find(b'=')
            index = len(kv) if index == -1 else index
            key = str(kv[:index], encoding='utf-8', errors='ignore')
            val = str(kv[index+1:], encoding='utf-8', errors='ignore')
            kvs[key] = val.replace('\\p', '|').replace('\\s', ' ').replace('\\/', '/')

        return kvs


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        teamspeak3 = Teamspeak3(host='145.239.200.2', port=10011, voice_port=9987, timeout=5.0)
        info = await teamspeak3.get_info()
        print(json.dumps(info, indent=None) + '\n')
        clients = await teamspeak3.get_clients()
        print(json.dumps(clients, indent=None) + '\n')
        channels = await teamspeak3.get_channels()
        print(json.dumps(channels, indent=None) + '\n')

    asyncio.run(main_async())



import aiohttp
import base64
import json

from opengsq.exceptions import ServerNotFoundException
from opengsq.protocol_base import ProtocolBase


class EOS(ProtocolBase):
    """Epic Online Services (EOS) Protocol"""
    full_name = 'Epic Online Services (EOS) Protocol'

    _api_url = 'https://api.epicgames.dev'

    def __init__(self, host: str, port: int, timeout: float = 5, client_id: str = None, client_secret: str = None, deployment_id: str = None):
        super().__init__(host, port, timeout)

        if client_id is None or client_secret is None or deployment_id is None:
            raise ValueError(
                "client_id, client_secret, and deployment_id must not be None")

        self.client_id = client_id
        self.client_secret = client_secret
        self.deployment_id = deployment_id
        self.access_token = None

    async def _get_access_token(self) -> str:
        url = f'{self._api_url}/auth/v1/oauth/token'
        body = f"grant_type=client_credentials&deployment_id={self.deployment_id}"
        headers = {
            "Authorization": f"Basic {base64.b64encode(f'{self.client_id}:{self.client_secret}'.encode('utf-8')).decode('utf-8')}",
            "Content-Type": "application/x-www-form-urlencoded"
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=body, headers=headers) as response:
                response.raise_for_status()
                data = await response.json()

        return data["access_token"]

    async def _get_matchmaking(self, data: dict):
        if self.access_token is None:
            self.access_token = await self._get_access_token()
            assert self.access_token is not None, "Failed to get access token"

        url = f"{self._api_url}/matchmaking/v1/{self.deployment_id}/filter"
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.access_token}"
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=json.dumps(data), headers=headers) as response:
                response.raise_for_status()
                data = await response.json()

        return data

    async def get_info(self) -> dict:
        data = await self._get_matchmaking({
            "criteria": [
                {
                    "key": "attributes.ADDRESS_s",
                    "op": "EQUAL",
                    "value": self._host
                },
                {
                    "key": "attributes.ADDRESSBOUND_s",
                    "op": "CONTAINS",
                    "value": f':{self._port}'
                },
            ]
        })

        if data["count"] <= 0:
            raise ServerNotFoundException()

        return data['sessions'][0]


if __name__ == '__main__':
    import asyncio

    async def main_async():
        client_id = 'xyza7891muomRmynIIHaJB9COBKkwj6n'
        client_secret = 'PP5UGxysEieNfSrEicaD1N2Bb3TdXuD7xHYcsdUHZ7s'
        deployment_id = 'ad9a8feffb3b4b2ca315546f038c3ae2'

        eos = EOS(host='150.138.77.118', port=10019, timeout=5.0, client_id=client_id,
                  client_secret=client_secret, deployment_id=deployment_id)
        data = await eos.get_info()
        print(json.dumps(data, indent=None) + '\n')

    asyncio.run(main_async())



from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class ASE(ProtocolBase):
    """All-Seeing Eye Protocol"""
    full_name = 'All-Seeing Eye Protocol'

    _request = b's'
    _response = b'EYE1'

    async def get_status(self) -> dict:
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send Request
            sock.send(self._request)

            # Server response
            response = await sock.recv()

        header = response[:4]

        if header != self._response:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self._response))
            )

        br = BinaryReader(response[4:])

        result = {}
        result['gamename'] = self.__read_string(br)
        result['gameport'] = self.__read_string(br)
        result['hostname'] = self.__read_string(br)
        result['gametype'] = self.__read_string(br)
        result['map'] = self.__read_string(br)
        result['version'] = self.__read_string(br)
        result['password'] = self.__read_string(br)
        result['numplayers'] = self.__read_string(br)
        result['maxplayers'] = self.__read_string(br)
        result['rules'] = self.__parse_rules(br)
        result['players'] = self.__parse_players(br)

        return result

    def __parse_rules(self, br: BinaryReader):
        rules = {}

        while not br.is_end():
            key = self.__read_string(br)

            if not key:
                break

            rules[key] = self.__read_string(br)

        return rules

    def __parse_players(self, br: BinaryReader):
        players = []
        keys = {1: 'name', 2: 'team', 4: 'skin', 8: 'score', 16: 'ping', 32: 'time'}

        while not br.is_end():
            flags = br.read_byte()
            player = {}

            for key, value in keys.items():
                if flags & key == key:
                    player[value] = self.__read_string(br)

            players.append(player)

        return players

    def __read_string(self, br: BinaryReader):
        length = br.read_byte()
        return str(br.read_bytes(length - 1), encoding='utf-8', errors='ignore')


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        # mtasa
        ase = ASE(host='79.137.97.3', port=22126, timeout=10.0)
        status = await ase.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



import argparse
import asyncio
import inspect
import json
import os
import re
import sys
from functools import partial
from pydoc import locate
from typing import Mapping, Sequence

from opengsq.protocol_base import ProtocolBase
from opengsq.version import __version__


class CLI:
    def __init__(self):
        self.__paths = {}

    def register(self, parser: argparse.ArgumentParser):
        # Add version argument
        parser.add_argument('-V', '--version', action='store_true', help='print the opengsq version number and exit')

        opengsq_path = os.path.abspath(os.path.dirname(__file__))
        subparsers = parser.add_subparsers(dest='subparser_name')
        pattern = re.compile(r'from\s+(\S+)\s+import\s+(.+,?\S)')

        # Load all protocols from __init__.py
        with open(os.path.join(opengsq_path, 'protocols', '__init__.py')) as f:
            for (protocol_path, protocol_classnames) in re.findall(pattern, f.read()):
                for protocol_classname in protocol_classnames.split(','):
                    name, fullpath, parameters = self.__extract(protocol_path, protocol_classname)

                    # Save to self.__paths dictionary
                    # Example: name = 'source', fullpath = 'opengsq.protocols.source.Source'
                    self.__paths[name] = fullpath

                    # Add parser and arguments
                    obj: ProtocolBase = locate(fullpath)
                    sub = subparsers.add_parser(name, help=obj.full_name)
                    self.__add_arguments(sub, parameters)
                    method_names = [func for func in dir(obj) if callable(getattr(obj, func)) and func.startswith('get_')]
                    sub.add_argument('--function', default=method_names[0], type=str, help='(default: %(default)s)')
                    sub.add_argument('--indent', default=None, type=int, nargs='?')

    # Get the query response in json format
    async def run(self, args: Sequence[str]) -> str:
        # Return version if -V or --version
        if args.version:
            return 'v' + __version__
        else:
            del args.version

        # Load the obj from path
        obj = locate(self.__paths[args.subparser_name])
        del args.subparser_name

        function = args.function
        del args.function

        indent = args.indent
        del args.indent

        # Bind values to obj parameters
        for value in vars(args).values():
            obj = partial(obj, value)

        # Create obj()
        protocol: ProtocolBase = obj()
        func = getattr(protocol, function)

        return json.dumps(await func(), ensure_ascii=False, indent=indent)

    # Extract name, fullpath, parameters from path, classname
    def __extract(self, path: str, classname: str):
        name = path.split('.')[-1]
        fullpath = '{}.{}'.format(path, classname.strip())
        parameters = inspect.signature(locate(fullpath).__init__).parameters

        return name, fullpath, parameters

    def __add_arguments(self, sub: argparse.ArgumentParser, parameters: Mapping[str, inspect.Parameter]):
        for key in parameters:
            if parameters[key].name == 'self':
                continue

            name_or_flags = '--{}'.format(parameters[key].name)
            required = parameters[key].default == inspect._empty
            default = None if required else parameters[key].default
            type = parameters[key].annotation
            help = None if required else '(default: %(default)s)'

            sub.add_argument(name_or_flags, default=default, required=required, type=type, help=help)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_async())
    loop.close()


async def main_async():
    cli = CLI()

    parser = argparse.ArgumentParser()
    cli.register(parser)

    if len(sys.argv) <= 1:
        parser.print_help(sys.stderr)
        sys.exit(-1)

    try:
        args = parser.parse_args()
        result = await cli.run(args)
        sys.stdout.write(result)
    except asyncio.exceptions.TimeoutError:
        sys.stderr.write('opengsq: error: timed out\n')
        sys.exit(-2)

    sys.exit(0)


if __name__ == '__main__':
    main()



import re

from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Quake1(ProtocolBase):
    """Quake1 Protocol"""
    full_name = 'Quake1 Protocol'

    def __init__(self, host: str, port: int, timeout: float = 5.0):
        """Quake1 Query Protocol"""
        super().__init__(host, port, timeout)
        self._delimiter1 = b'\\'
        self._delimiter2 = b'\n'
        self._request_header = b'status'
        self._response_header = 'n'

    async def get_status(self) -> dict:
        """This returns server information and players."""
        br = await self._get_response_binary_reader()

        return {
            'info': self._parse_info(br),
            'players': self._parse_players(br),
        }

    async def _get_response_binary_reader(self) -> BinaryReader:
        response_data = await self._connect_and_send(self._request_header)

        br = BinaryReader(response_data)
        header = br.read_string(self._delimiter1)

        if header != self._response_header:
            raise Exception(f'Packet header mismatch. Received: {header}. Expected: {self._response_header}.')

        return br

    def _parse_info(self, br: BinaryReader) -> dict:
        info = {}

        # Read all key values until meet \n
        while br.length() > 0:
            key = br.read_string(self._delimiter1)

            if key == '':
                break

            info[key] = br.read_string([self._delimiter1, self._delimiter2])

            br.stream_position -= 1

            if bytes([br.read_byte()]) == self._delimiter2:
                break

        return info

    def _parse_players(self, br: BinaryReader) -> list:
        players = []

        for matches in self._get_player_match_collections(br):
            matches: list[re.Match] = [match.group() for match in matches]

            players.append({
                'id': int(matches[0]),
                'score': int(matches[1]),
                'time': int(matches[2]),
                'ping': int(matches[3]),
                'name': str(matches[4]).strip('"'),
                'skin': str(matches[5]).strip('"'),
                'color1': int(matches[6]),
                'color2': int(matches[7]),
            })

        return players

    def _get_player_match_collections(self, br: BinaryReader):
        match_collections = []

        # Regex to split with whitespace and double quote
        regex = re.compile(r'"(\\"|[^"])*?"|[^\s]+')

        # Read all players
        while not br.is_end():
            match_collections.append(regex.finditer(br.read_string(self._delimiter2)))

        return match_collections

    async def _connect_and_send(self, data):
        # Connect to remote host
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            header = b'\xFF\xFF\xFF\xFF'

            # Send Request
            sock.send(header + data + b'\x00')

            # Server response
            response_data = await sock.recv()

        # Remove the last 0x00 if exists (Only if Quake1)
        if response_data[-1] == 0:
            response_data = response_data[:-1]

        # Add \n at the last of responseData if not exists
        if response_data[-1] != self._delimiter2:
            response_data += self._delimiter2

        # Remove the first four 0xFF
        return response_data[len(header):]


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        quake1 = Quake1(host='35.185.44.174', port=27500, timeout=5.0)
        status = await quake1.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



# -*- coding: utf-8 -*-
# flake8: noqa

"""
OpenGSQ Python Library
~~~~~~~~~~~~~~~~~~~

Python library for querying game servers

:copyright: (c) 2021 BattlefieldDuck
:license: MIT, see LICENSE for more details.

"""

from opengsq.protocols import *



import json
import re
import struct

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync, SocketKind


class Minecraft(ProtocolBase):
    """Minecraft Protocol (https://wiki.vg/Server_List_Ping)"""
    full_name = 'Minecraft Protocol'

    async def get_status(self, version=47, strip_color=True) -> dict:
        """Get Status

        Args:
            version (int, optional): https://wiki.vg/Protocol_version_numbers. Defaults to 47.
            strip_color (bool, optional): Strip color. Defaults to True.

        Returns:
            dict: status dict
        """

        # Prepare the request
        address = self._host.encode('utf8')
        protocol = self._pack_varint(version)
        request = b'\x00' + protocol + self._pack_varint(len(address)) + address + struct.pack('H', self._port) + b'\x01'
        request = self._pack_varint(len(request)) + request + b'\x01\x00'

        with SocketAsync(SocketKind.SOCK_STREAM) as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))
            sock.send(request)

            response = await sock.recv()
            br = BinaryReader(response)
            length = self._unpack_varint(br)

            # Keep recv() until reach packet length
            while len(response) < length:
                response += await sock.recv()

        # Read fill response
        br = BinaryReader(response)
        self._unpack_varint(br)  # packet length
        self._unpack_varint(br)  # packet id
        count = self._unpack_varint(br)  # json length

        # The packet may response with two json objects, so we need to get the json length exactly
        data = json.loads(br.read_bytes(count).decode('utf-8'))

        if strip_color:
            if 'sample' in data['players']:
                for i, player in enumerate(data['players']['sample']):
                    data['players']['sample'][i]['name'] = Minecraft.strip_colors(player['name'])

            if isinstance(data['description'], str):
                data['description'] = Minecraft.strip_colors(data['description'])

            if 'text' in data['description'] and isinstance(data['description']['text'], str):
                data['description']['text'] = Minecraft.strip_colors(data['description']['text'])

            if 'extra' in data['description'] and isinstance(data['description']['extra'], list):
                for i, extra in enumerate(data['description']['extra']):
                    if isinstance(extra['text'], str):
                        data['description']['extra'][i]['text'] = Minecraft.strip_colors(extra['text'])

        return data

    async def get_status_pre17(self, strip_color=True) -> dict:
        """Get ping info from a server that uses a version older than Minecraft 1.7"""
        with SocketAsync(SocketKind.SOCK_STREAM) as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))
            sock.send(b'\xFE\x01')
            response = await sock.recv()

        br = BinaryReader(response)
        header = br.read_byte()

        if header != 0xFF:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(0xFF))
            )

        br.read_bytes(2)  # length of the following string
        items = br.read().split(b'\x00\x00')

        result = {}
        result['protocol'] = str(items[1], 'utf-16be')
        result['version'] = str(items[2], 'utf-16be')
        result['motd'] = str(items[3], 'utf-16be')
        result['numplayers'] = int(str(items[4], 'utf-16be'))
        result['maxplayers'] = int(str(items[5], 'utf-16be'))

        if strip_color:
            result['motd'] = Minecraft.strip_colors(result['motd'])

        return result

    @staticmethod
    def strip_colors(text: str):
        """Strip color codes"""
        return re.compile(r'\u00A7[0-9A-FK-OR]', re.IGNORECASE).sub('', text)

    def _pack_varint(self, val: int):
        total = b''

        if val < 0:
            val = (1 << 32) + val

        while val >= 0x80:
            bits = val & 0x7F
            val >>= 7
            total += struct.pack('B', (0x80 | bits))

        bits = val & 0x7F
        total += struct.pack('B', bits)

        return total

    def _unpack_varint(self, br: BinaryReader):
        total = 0
        shift = 0
        val = 0x80

        while val & 0x80:
            val = struct.unpack('B', br.read_bytes(1))[0]
            total |= ((val & 0x7F) << shift)
            shift += 7

        if total & (1 << 31):
            total -= (1 << 32)

        return total


if __name__ == '__main__':
    import asyncio

    async def main_async():
        minecraft = Minecraft(host='valistar.site', port=25565, timeout=5.0)
        status = await minecraft.get_status(47, strip_color=True)
        print(json.dumps(status, indent=None, ensure_ascii=False) + '\n')
        status = await minecraft.get_status_pre17()
        print(json.dumps(status, indent=None, ensure_ascii=False) + '\n')

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Unreal2(ProtocolBase):
    """Unreal 2 Protocol"""
    full_name = 'Unreal 2 Protocol'

    _DETAILS = 0x00
    _RULES = 0x01
    _PLAYERS = 0x02

    async def get_details(self):
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send Request
            sock.send(b'\x79\x00\x00\x00' + bytes([self._DETAILS]))

            # Server response
            response = await sock.recv()

        # Remove the first 4 bytes \x80\x00\x00\x00
        br = BinaryReader(response[4:])
        header = br.read_byte()

        if header != self._DETAILS:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self._DETAILS))
            )

        details = {}
        details['ServerId'] = br.read_long()  # 0
        details['ServerIP'] = br.read_string()  # empty
        details['GamePort'] = br.read_long()
        details['QueryPort'] = br.read_long()  # 0
        details['ServerName'] = self.__read_string(br)
        details['MapName'] = self.__read_string(br)
        details['GameType'] = self.__read_string(br)
        details['NumPlayers'] = br.read_long()
        details['MaxPlayers'] = br.read_long()

        if br.length() > 12:
            try:
                # Killing Floor
                stream_position = br.stream_position
                details['WaveCurrent'] = br.read_long()
                details['WaveTotal'] = br.read_long()
                details['Ping'] = br.read_long()
                details['Flags'] = br.read_long()
                details['Skill'] = self.__read_string(br)
            except Exception:
                br.stream_position = stream_position
                details['Ping'] = br.read_long()
                details['Flags'] = br.read_long()
                details['Skill'] = self.__read_string(br)

        return details

    async def get_rules(self):
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send Request
            sock.send(b'\x79\x00\x00\x00' + bytes([self._RULES]))

            # Server response
            response = await sock.recv()

        # Remove the first 4 bytes \x80\x00\x00\x00
        br = BinaryReader(response[4:])
        header = br.read_byte()

        if header != self._RULES:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self._RULES))
            )

        rules = {}
        rules['Mutators'] = []

        while not br.is_end():
            key = self.__read_string(br)
            val = self.__read_string(br)

            if key.lower() == 'mutator':
                rules['Mutators'].append(val)
            else:
                rules[key] = val

        return rules

    async def get_players(self):
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send Request
            sock.send(b'\x79\x00\x00\x00' + bytes([self._PLAYERS]))

            # Server response
            response = await sock.recv()

        # Remove the first 4 bytes \x80\x00\x00\x00
        br = BinaryReader(response[4:])
        header = br.read_byte()

        if header != self._PLAYERS:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self._PLAYERS))
            )

        players = []

        while not br.is_end():
            player = {}
            player['Id'] = br.read_long()
            player['Name'] = self.__read_string(br)
            player['Ping'] = br.read_long()
            player['Score'] = br.read_long()
            player['StatsId'] = br.read_long()
            players.append(player)

        return players

    @staticmethod
    def strip_colors(text: bytes):
        """Strip color codes"""
        string = re.compile(b'\x7f|[\x00-\x1a]|[\x1c-\x1f]').sub(b'', text)
        return string.replace(b'\x1b@@', b'').replace(b'\x1b@', b'').replace(b'\x1b', b'')

    def __read_string(self, br: BinaryReader):
        length = br.read_byte()
        string = br.read_string()

        if length == len(string) + 1:
            b = bytes(string, encoding='utf-8')
        else:
            b = bytes(string, encoding='utf-16')

        b = Unreal2.strip_colors(b)

        return b.decode('utf-8', 'ignore')


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        # ut2004
        unreal2 = Unreal2(host='109.230.224.189', port=6970, timeout=10.0)
        details = await unreal2.get_details()
        print(json.dumps(details, indent=None) + '\n')
        rules = await unreal2.get_rules()
        print(json.dumps(rules, indent=None) + '\n')
        players = await unreal2.get_players()
        print(json.dumps(players, indent=None) + '\n')

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Doom3(ProtocolBase):
    """Doom3 Protocol"""
    full_name = 'Doom3 Protocol'

    _player_fields = {
        'doom': ['id', 'ping', 'rate', 'name'],
        'quake4': ['id', 'ping', 'rate', 'name', 'clantag'],
        'etqw': ['id', 'ping', 'name', 'clantag_pos', 'clantag', 'typeflag']
    }

    async def get_info(self, strip_color=True):
        request = b'\xFF\xFFgetInfo\x00ogsq\x00'
        response = await SocketAsync.send_and_receive(self._host, self._port, self._timeout, request)

        # Remove the first two 0xFF
        br = BinaryReader(response[2:])
        header = br.read_string()

        if header != 'infoResponse':
            raise InvalidPacketException(f'Packet header mismatch. Received: {header}. Expected: infoResponse.')

        # Read challenge
        br.read_bytes(4)

        if br.read_bytes(4) != b'\xff\xff\xff\xff':
            br.stream_position -= 4

        info = {}

        # Read protocol version
        minor = br.read_short()
        major = br.read_short()
        info['version'] = f'{major}.{minor}'

        # Read packet size
        if br.read_long() != br.length():
            br.stream_position -= 4

        # Key / value pairs, delimited by an empty pair
        while br.length() > 0:
            key = br.read_string().strip()
            val = br.read_string().strip()

            if key == '' and val == '':
                break

            info[key] = Doom3.strip_colors(val) if strip_color else val

        stream_position = br.stream_position

        # Try parse the fields
        for mod in self._player_fields.keys():
            try:
                info['players'] = self.__parse_player(br, self._player_fields[mod], strip_color)
                break
            except Exception:
                info['players'] = []
                br.stream_position = stream_position

        return info

    def __parse_player(self, br: BinaryReader, fields: list, strip_color: bool):
        players = []

        def value_by_field(field: str, br: BinaryReader):
            if field == 'id' or field == 'clantag_pos' or field == 'typeflag':
                return br.read_byte()
            elif field == 'ping':
                return br.read_short()
            elif field == 'rate':
                return br.read_long()

            string = br.read_string()

            return Doom3.strip_colors(string) if strip_color else string

        while True:
            player = {field: value_by_field(field, br) for field in fields}

            if player['id'] == 32:
                break

            players.append(player)

        return players

    @staticmethod
    def strip_colors(text: str):
        """Strip color codes"""
        return re.compile('\\^(X.{6}|.)').sub('', text)


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        # doom3
        doom3 = Doom3(host='66.85.14.240', port=27666, timeout=5.0)
        info = await doom3.get_info()
        print(json.dumps(info, indent=None) + '\n')

        # etqw
        doom3 = Doom3(host='178.162.135.83', port=27735, timeout=5.0)
        info = await doom3.get_info()
        print(json.dumps(info, indent=None) + '\n')

        # quake4
        doom3 = Doom3(host='88.99.0.7', port=28007, timeout=5.0)
        info = await doom3.get_info()
        print(json.dumps(info, indent=None) + '\n')

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class GameSpy3(ProtocolBase):
    """GameSpy Protocol version 3"""
    full_name = 'GameSpy Protocol version 3'
    challenge = False

    async def get_status(self):
        """Retrieves information about the server including, Info, Players, and Teams."""
        # Connect to remote host
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            request_h = b'\xFE\xFD'
            timestamp = b'\x04\x05\x06\x07'
            challenge = b''

            if self.challenge:
                # Packet 1: Initial request - (https://wiki.unrealadmin.org/UT3_query_protocol#Packet_1:_Initial_request)
                sock.send(request_h + b'\x09' + timestamp)

                # Packet 2: First response - (https://wiki.unrealadmin.org/UT3_query_protocol#Packet_2:_First_response)
                response = await sock.recv()

                if response[0] != 9:
                    raise InvalidPacketException(
                        'Packet header mismatch. Received: {}. Expected: {}.'
                        .format(chr(response[0]), chr(9))
                    )

                # Packet 3: Second request - (http://wiki.unrealadmin.org/UT3_query_protocol#Packet_3:_Second_request)
                challenge = int(response[5:].decode('ascii').strip('\x00'))
                challenge = b'' if challenge == 0 else challenge.to_bytes(4, 'big', signed=True)

            request_data = request_h + b'\x00' + timestamp + challenge
            sock.send(request_data + b'\xFF\xFF\xFF\x01')

            # Packet 4: Server information response
            # (http://wiki.unrealadmin.org/UT3_query_protocol#Packet_4:_Server_information_response)
            response = await self.__read(sock)

        br = BinaryReader(response)

        result = {}
        result['info'] = {}

        while True:
            key = br.read_string()

            if key == '':
                break

            result['info'][key] = br.read_string()

        pattern = re.compile(rb'\x00([^a-zA-Z])([a-zA-Z_]+)\x00\x00(.+?(?=\x00\x00))')
        current_id = -1
        current_name = None

        for (id, name, data) in re.findall(pattern, b'\x00' + br.read()):
            values = data.split(b'\x00')
            name = name.decode('utf-8').split('_')[0]

            if current_id != id and id != b'\x00':
                current_id, current_name = id, name
                result[current_name] = [{} for _ in range(len(values))]

            for i in range(len(values)):
                result[current_name][i][name] = values[i].decode('utf-8')

        return result

    async def __read(self, sock) -> bytes:
        packet_count = -1
        payloads = {}

        while packet_count == -1 or len(payloads) > packet_count:
            response = await sock.recv()

            br = BinaryReader(response)
            header = br.read_byte()

            if header != 0:
                raise InvalidPacketException(
                    'Packet header mismatch. Received: {}. Expected: {}.'
                    .format(chr(header), chr(0))
                )

            # Skip the timestamp and splitnum
            br.read_bytes(13)

            # The 'numPackets' byte
            num_packets = br.read_byte()

            # The low 7 bits are the packet index (starting at zero)
            number = num_packets & 0x7F

            # The high bit is whether or not this is the last packet
            if num_packets & 0x80:
                # Set packet_count if it is the last packet
                packet_count = number + 1

            # The object id
            # 0 = server kv information
            # 1 = player_   \x00\x01player_\x00\x00 since \x01
            # 2 = team_t    \x00\x02team_t\x00\x00  since \x02
            # etc...
            obj_id = br.read_byte()
            header = b''

            if obj_id >= 1:
                # The object key name
                string = br.read_string()

                # How many times did the value appear in the previous packet
                count = br.read_byte()

                # Set back the packet header if it didn't appear before
                header = b'\x00' + bytes([obj_id]) + string.encode() + b'\x00\x00' if count == 0 else b''

            payload = header + br.read()[:-1]

            # Remove the last trash string on the payload
            payloads[number] = payload[:payload.rfind(b'\x00') + 1]

        response = b''.join(payloads[number] for number in sorted(payloads))

        return response


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        gs3 = GameSpy3(host='185.107.96.59', port=29900, timeout=5.0)
        server = await gs3.get_status()
        print(json.dumps(server, indent=None))

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class GameSpy1(ProtocolBase):
    """GameSpy Protocol version 1"""
    full_name = 'GameSpy Protocol version 1'

    # Legacy:UT_Server_Query - (https://wiki.beyondunreal.com/Legacy:UT_Server_Query)
    # Query_commands - (https://wiki.beyondunreal.com/XServerQuery#Query_commands)
    class __Request():
        BASIC = b'\\basic\\'
        INFO = b'\\info\\xserverquery'
        RULES = b'\\rules\\xserverquery'
        PLAYERS = b'\\players\\xserverquery'
        STATUS = b'\\status\\xserverquery'
        TEAMS = b'\\teams\\'

    async def get_basic(self) -> dict:
        """This returns basic server information, mainly for recognition."""
        return self.__parse_as_key_values(await self.__connect_and_send(self.__Request.BASIC))

    # Server may still response with Legacy version
    async def get_info(self, xserverquery: bool = True) -> dict:
        """
        Information about the current game running on the server.

        If the server uses XServerQuery, he sends you the new information, otherwise he'll give you back the old information.
        """
        data = xserverquery and self.__Request.INFO or self.__Request.INFO.replace(b'xserverquery', b'')
        return self.__parse_as_key_values(await self.__connect_and_send(data))

    async def get_rules(self, xserverquery: bool = True) -> list:
        """
        Setting for the current game, return sets of rules depends on the running game type.

        If the server uses XServerQuery, he sends you the new information, otherwise he'll give you back the old information.
        """
        data = xserverquery and self.__Request.RULES or self.__Request.RULES.replace(b'xserverquery', b'')
        return self.__parse_as_key_values(await self.__connect_and_send(data))

    async def get_players(self, xserverquery: bool = True) -> list:
        """
        Returns information about each player on the server.

        If the server uses XServerQuery, he sends you the new information, otherwise he'll give you back the old information.
        """
        data = xserverquery and self.__Request.PLAYERS or self.__Request.PLAYERS.replace(b'xserverquery', b'')
        return self.__parse_as_object(await self.__connect_and_send(data))

    async def get_status(self, xserverquery: bool = True) -> dict:
        """
        XServerQuery: \\info\\xserverquery\\rules\\xserverquery\\players\\xserverquery

        Old response: \\basic\\info\\rules\\players\\

        If the server uses XServerQuery, he sends you the new information, otherwise he'll give you back the old information.
        """
        data = xserverquery and self.__Request.STATUS or self.__Request.STATUS.replace(b'xserverquery', b'')
        br = await self.__connect_and_send(data)

        status = {}
        status['info'] = self.__parse_as_key_values(br, is_status=True)
        status['players'] = self.__parse_as_object(br, is_player=True)
        status['teams'] = [] if br.is_end() else self.__parse_as_object(br)

        return status

    async def get_teams(self) -> list:
        """Returns information about each team on the server."""
        return self.__parse_as_object(await self.__connect_and_send(self.__Request.TEAMS))

    # Receive packets and sort it
    async def __get_packets_response(self, sock: SocketAsync):
        payloads = {}
        packet_count = -1

        # Loop until received all packets
        while packet_count == -1 or len(payloads) < packet_count:
            packet = await sock.recv()

            # Get the packet number from query_id
            r = re.compile(rb'\\queryid\\\d+\.(\d+)')
            number, payload = int(r.search(packet).group(1)), r.sub(b'', packet)

            # If it is the last packet, it will contain b'\\final\\' at the end of the response
            if payload.endswith(b'\\final\\'):
                # Save the packet count
                packet_count = number

                # Remove the last b'\\final\\'
                payload = payload[:-7]

            # Save the payload, remove the first byte if it is the first packet
            payloads[number] = payload[1:] if number == 1 else payload

        # Sort the payload and return as bytes
        response = b''.join(payloads[number] for number in sorted(payloads))

        return response

    async def __connect_and_send(self, data) -> BinaryReader:
        # Connect to remote host
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            sock.send(data)
            br = BinaryReader(await self.__get_packets_response(sock))

        return br

    def __parse_as_key_values(self, br: BinaryReader, is_status=False):
        kv = {}

        # Bind key value
        while br.length() > 0:
            key = br.read_string(b'\\').lower()

            # Check is the end of key_values
            if is_status:
                items = key.split('_')

                if len(items) > 1 and items[1].isdigit():
                    # Read already, so add it back
                    br.prepend_bytes(key.encode() + b'\\')
                    break

            value = br.read_string(b'\\')
            kv[key] = value.strip()

        return kv

    def __parse_as_object(self, br: BinaryReader, is_player=False):
        items, keyhashes, filters = {}, [], []

        while br.length() > 0:
            # Get the key, for example player_1, frags_1, ping_1, etc...
            key = br.read_string(b'\\').lower()

            if is_player and key.startswith('teamname_'):
                # Read already, so add it back
                br.prepend_bytes(key.encode() + b'\\')
                break

            # Extract to name and index, for example name=player, index=1
            matches = re.search(r'(.+?)_(\d+)', key)
            name = matches.group(1)
            name = 'player' if name == 'playername' else name
            name = 'team' if name == 'teamname' else name
            index = int(matches.group(2))

            # Get the value, and strip it since some values contain whitespaces
            value = br.read_string(b'\\').strip()

            # Some servers (bf1942) report the same player multiple times, so filter it by keyhash
            if name == 'keyhash':
                if value in keyhashes:
                    filters.append(index)
                else:
                    keyhashes.append(value)

            # Create a dict on items if next index appears
            if index not in items:
                items[index] = {}

            # Save
            items[index][name] = value

        # Filter items by filters
        return [v for k, v in items.items() if k not in filters]


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        gs1 = GameSpy1(host='51.81.48.224', port=23000, timeout=5.0)  # bfield1942
        # gs1 = GameSpy1(address='139.162.235.20', query_port=7778, timeout=5.0)  # ut
        # gs1 = GameSpy1(address='192.223.24.6', query_port=7778, timeout=5.0)  # ut
        gs1 = GameSpy1(host='141.94.205.35', port=12300, timeout=5.0)  # mohaa
        status = await gs1.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync, SocketKind


class Battlefield(ProtocolBase):
    """Battlefield Protocol"""
    full_name = 'Battlefield Protocol'

    _info = b'\x00\x00\x00\x21\x1b\x00\x00\x00\x01\x00\x00\x00\x0a\x00\x00\x00serverInfo\x00'
    _version = b'\x00\x00\x00\x22\x18\x00\x00\x00\x01\x00\x00\x00\x07\x00\x00\x00version\x00'
    _players = b'\x00\x00\x00\x23\x24\x00\x00\x00\x02\x00\x00\x00\x0b\x00\x00\x00listPlayers\x00\x03\x00\x00\x00all\x00'

    async def get_info(self) -> dict:
        data = await self.__get_data(self._info)

        info = {}
        info['hostname'] = str(data.pop(0)).strip()
        info['numplayers'] = int(data.pop(0))
        info['maxplayers'] = int(data.pop(0))
        info['gametype'] = data.pop(0)
        info['map'] = data.pop(0)
        info['roundsplayed'] = int(data.pop(0))
        info['roundstotal'] = int(data.pop(0))
        num_teams = int(data.pop(0))
        info['teams'] = [float(data.pop(0)) for _ in range(num_teams)]
        info['targetscore'] = int(data.pop(0))
        info['status'] = data.pop(0)
        info['ranked'] = data.pop(0) == 'true'
        info['punkbuster'] = data.pop(0) == 'true'
        info['password'] = data.pop(0) == 'true'
        info['uptime'] = int(data.pop(0))
        info['roundtime'] = int(data.pop(0))

        try:
            if data[0] == 'BC2':
                info['mod'] = data.pop(0)
                data.pop(0)

            info['ip_port'] = data.pop(0)
            info['punkbuster_version'] = data.pop(0)
            info['join_queue'] = data.pop(0) == 'true'
            info['region'] = data.pop(0)
            info['pingsite'] = data.pop(0)
            info['country'] = data.pop(0)

            try:
                info['blaze_player_count'] = int(data[0])
                info['blaze_game_state'] = data[1]
            except Exception:
                info['quickmatch'] = data.pop(0) == 'true'
        except Exception:
            pass

        return info

    async def get_version(self) -> dict:
        data = await self.__get_data(self._version)
        return {'mod': data[0], 'version': data[1]}

    async def get_players(self) -> list:
        data = await self.__get_data(self._players)
        count = int(data.pop(0))  # field count
        fields, data = data[:count], data[count:]
        numplayers = int(data.pop(0))
        players = []

        for _ in range(numplayers):
            values, data = data[:count], data[count:]
            players.append(dict(zip(fields, values)))

        return players

    async def __get_data(self, request: bytes):
        kind = SocketKind.SOCK_STREAM
        response = await SocketAsync.send_and_receive(self._host, self._port, self._timeout, request, kind)
        return self.__decode(response)

    def __decode(self, response: bytes):
        br = BinaryReader(response)
        br.read_long()  # header
        br.read_long()  # packet length
        count = br.read_long()  # string count
        data = []

        for _ in range(count):
            br.read_long()  # length of the string
            data.append(br.read_string())

        return data[1:]


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        entries = [
            ('91.206.15.69', 48888),  # bfbc2
            ('94.250.199.214', 47200),  # bf3
            ('74.91.124.140', 47200),  # bf4
            ('185.189.255.240', 47600),  # bfh
        ]

        for address, query_port in entries:
            battlefield = Battlefield(address, query_port, timeout=10.0)
            info = await battlefield.get_info()
            print(json.dumps(info, indent=None) + '\n')
            version = await battlefield.get_version()
            print(json.dumps(version, indent=None) + '\n')
            players = await battlefield.get_players()
            print(json.dumps(players, indent=None) + '\n')

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.protocols.quake2 import Quake2


class Quake3(Quake2):
    """Quake3 Protocol"""
    full_name = 'Quake3 Protocol'

    def __init__(self, host: str, port: int, timeout: float = 5.0):
        """Quake3 Query Protocol"""
        super().__init__(host, port, timeout)
        self._request_header = b'getstatus'
        self._response_header = 'statusResponse\n'

    async def get_info(self, strip_color=True) -> dict:
        """This returns server information only."""
        response_data = await self._connect_and_send(b'getinfo opengsq')

        br = BinaryReader(response_data)
        header = br.read_string(self._delimiter1)
        response_header = 'infoResponse\n'

        if header != response_header:
            raise Exception(f'Packet header mismatch. Received: {header}. Expected: {response_header}.')

        info = self._parse_info(br)

        if not strip_color:
            return info

        if 'hostname' in info:
            info['hostname'] = Quake3.strip_colors(info['hostname'])

        return info

    async def get_status(self, strip_color=True) -> dict:
        """This returns server information and players."""
        br = await self._get_response_binary_reader()

        status = {
            'info': self._parse_info(br),
            'players': self._parse_players(br),
        }

        if not strip_color:
            return status

        if 'sv_hostname' in status['info']:
            status['info']['sv_hostname'] = Quake3.strip_colors(status['info']['sv_hostname'])

        for player in status['players']:
            if 'name' in player:
                player['name'] = Quake3.strip_colors(player['name'])

        return status

    @staticmethod
    def strip_colors(text: str):
        """Strip color codes"""
        return re.compile('\\^(X.{6}|.)').sub('', text)


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        quake3 = Quake3(host='85.10.197.106', port=27960, timeout=5.0)
        info = await quake3.get_info()
        status = await quake3.get_status()
        print(json.dumps(info, indent=None) + '\n')
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())
